module prop2dutils

	use wilson
	use dwilson
	use helicon
	use dhelicon
	use chelix
	use ccrosskolmog
	implicit none

contains

	real function SrcFunc (fmax, t, tshift)
		real, intent(in)	:: t, tshift
		real, intent(in)	:: fmax

		SrcFunc = exp (-0.5*fmax*fmax*(t-tshift)*(t-tshift)) * cos (3.141592*fmax*(t-tshift))

	end function


	real function GetVmax (v, n1, n2)
		real, dimension(:)   :: v
		integer, intent (in) :: n1, n2
		integer				 :: i, j
		real				 :: vv

		vv = 0.

		do j = 1, n2
			do i = 1, n1
				if (v((j-1) * n1 + i) .gt. vv) then
					vv = v((j-1) * n1 + i)
				end if
			end do
		end do
	
		GetVmax = vv
	end function

	real function GetVavg (v, n1, n2)
		real, dimension(:)   :: v
		integer, intent (in) :: n1, n2
		integer				 :: i, j
		real(8)				 :: vsum
		real				 :: vavg

		vsum = 0.
		vavg = 0.

		do j = 1, n2
			do i = 1, n1
				vsum = vsum + v((j-1)*n1 + i)
			end do
		end do
		vavg = vsum / (n1 * n2)	

		GetVavg = vavg
	end function

	subroutine Pad2D (M, nz, padtop, padbottom, nx, padleft, padright)
		real, dimension(:,:)	:: M
		real, intent (in)		:: nz, nx, padtop, padbottom, padleft, padright
		integer					:: i, j

		do j = padleft, 1, -1 ! pad model horizontal left
			do i = padtop + 1, nz + padtop, 1
				M (i, j) = M (i, j+1)
			end do
		end do

		do j = nx + padleft + 1, nx + padleft + padright, 1 ! pad model horizontal right
			do i = padtop + 1, nz + padtop, 1
				M (i, j) = M (i, j-1)
			end do
		end do

		do i = padtop, 1, -1 ! pad model vertical top
			do j = 1, nx + padleft + padright, 1
				M (i, j) = M (i+1, j)
			end do
		end do

		do i = nz + padtop + 1, nz + padtop + padbottom, 1 ! pad model vertical bottom
			do j = 1, nx + padleft + padright, 1
				M (i, j) = M (i-1, j)
			end do
		end do
	end subroutine

	subroutine MkDampCoeff (Damp, nx, padleft, padright, nz, padtop, padbottom, b)
		real, dimension (:,:)	:: Damp
		integer, intent (in)	:: nx, nz, padleft, padright, padtop, padbottom
		real, intent (in)		:: b

		integer					:: i, j
		real					:: m_pi

		m_pi = 2 * acos(0.)

		Damp = 1.0

		do j = padleft, 1, -1
			Damp (:, j) = Damp (:, j) * (1 - b * (1 + cos (m_pi * j / padleft ) ))
		end do

		do j = nx + padleft + 1, nx + padleft + padright, 1
			Damp (:, j) = Damp (:, j) * (1 - b * (1 + cos (m_pi * (nz + padtop + padbottom - j) / padright) ))
		end do

		do i = padtop, 1, -1
			Damp (i, :) = Damp (i, :) * (1 - b * (1 + cos (m_pi * i / padtop)) )
		end do

		do i = nz + padtop + 1, nz + padtop + padbottom, 1
			Damp (i, :) = Damp (i, :) * (1 - b * (1 + cos (m_pi * (nz + padtop + padbottom - i) / padbottom) ))
		end do

	end subroutine


!!$=head1 NAME
!!$
!!$impfac -  factor a 2-D implicit 2nd order FD scheme
!!$
!!$=head1 SYNOPSIS
!!$
!!$C<aa= impfac2(eps,n1,na)>
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item eps - real
!!$
!!$      small value for approximation
!!$
!!$=item n1  - integer
!!$
!!$      length of n1 axis
!!$
!!$=item na - integer
!!$
!!$      1/2 number of filter coefs for representation
!!$
!!$=back
!!$
!!$=head1 RETURN VALUE
!!$
!!$=over 4
!!$
!!$=item aa    - type(filter)
!!$
!!$      Helix filter
!!$
!!$=back
!!$
!!$=head1  DESCRIPTION
!!$
!!$Create a one sided, minimum phase filter that approximates an implicit FD 2nd order scheme in 2D
!!$
!!$=head1 SEE ALSO
!!$
!!$L<wilson>
!!$
!!$=head1 LIBRARY
!!$
!!$B<geef90>
!!$
!$=cut
! Factor 2-D implicit FD 2nd spatial order scheme, using wilson-burg factorization.
	function impfac2 (eps, b, a1, a2, n1, na)   result (aa)
		type (filter)                       :: aa, imp
		real,                   intent (in) :: eps, b, a1, a2
		integer,                intent (in) :: n1, na
		integer                             :: i
		real                                :: a0, imp0

		call allocatehelix (imp, 2)	! implicit FD filter
		imp0       = b + eps		! zero lag coeff.
		imp%lag    = (/ 1, n1 /)	! lag(1)= 1; lag(2)=n1  # one side only
		imp%flt(1) = a1
		imp%flt(2) = a2

		call allocatehelix (aa, 2*na)	! implicit FD derivative
		aa%flt = 0.

		do i = 1, na  
			aa%lag (i     ) = i				! early lags (first row)
			aa%lag (na + i) = n1 + i - na	! late lags (second row)
		end do 

		call wilson_init	(10 * n1)
		call wilson_factor	(20, imp0, imp, a0, aa)
		call wilson_close	()
		call deallocatehelix(imp)
	end function 

! Factor 2-D implicit FD 4th spatial order scheme, using wilson-burg factorization.
	function impfac4 (eps, c, b1, b2, a1, a2, n1, na)   result (aa)
		type (filter)                       :: aa, imp
		real,                   intent (in) :: eps, c, b1, b2, a1, a2
		integer,                intent (in) :: n1, na
		integer                             :: i
		real                                :: a0, imp0

		call allocatehelix (imp, 4)	! implicit FD filter
		imp0       = c + eps		! zero lag coeff.
		imp%lag    = (/ 1, 2, n1, 2*n1 /)	! lag(1)= 1, lag(2)=2, lag(3)=n1, lag(4)=2*n1  # one side only
		imp%flt(1) = b1
		imp%flt(2) = a1
		imp%flt(3) = b2
		imp%flt(4) = a2

		call allocatehelix (aa, 4*na)	! implicit FD derivative
		aa%flt = 0.

		do i = 1, na  
			aa%lag (i     )   = i				! early lags (first row)
			aa%lag (na + i)   = n1 + i - na		! middle lags (second row)
			aa%lag (2*na + i) = n1 + i			! middle lags (second row)
			aa%lag (3*na + i) = 2*n1 + i - na	! late lags (third row)
		end do 

		call wilson_init	(10 * n1)
		call wilson_factor	(20, imp0, imp, a0, aa)
		call wilson_close	()
		call deallocatehelix(imp)
	end function 

! Factor 2-D implicit FD 4th spatial order scheme, double precision, using wilson-burg factorization.
	function dimpfac4 (eps, c, b1, b2, a1, a2, n1, na)   result (aa)
		type (dfilter)			:: aa, imp
		real(8), intent (in)	:: eps, c, b1, b2, a1, a2
		integer, intent (in)	:: n1, na
		integer					:: i
		real(8)					:: a0, imp0

		call dallocatehelix (imp, 4)	! implicit FD filter
		imp0       = c + eps			! zero lag coeff.
		imp%lag    = (/ 1, 2, n1, 2*n1 /)	! lag(1)= 1, lag(2)=2, lag(3)=n1, lag(4)=2*n1  # one side only
		imp%flt(1) = b1
		imp%flt(2) = a1
		imp%flt(3) = b2
		imp%flt(4) = a2

		call dallocatehelix (aa, 4*na)	! implicit FD derivative
		aa%flt = 0.

		do i = 1, na  
			aa%lag (i     )   = i				! early lags (first row)
			aa%lag (na + i)   = n1 + i - na		! middle lags (second row)
			aa%lag (2*na + i) = n1 + i			! middle lags (second row)
			aa%lag (3*na + i) = 2*n1 + i - na	! late lags (third row)
		end do 

		call dwilson_init	(10 * n1)
		call dwilson_factor	(20, imp0, imp, a0, aa)
		call dwilson_close	()
		call ddeallocatehelix (imp)
	end function 

! Factor 1-D implicit FD scheme.
	function impfac1d (eps, b, a1, n1, na)   result (aa)
		type (filter)                       :: aa, imp
		real,                   intent (in) :: eps, b, a1
		integer,                intent (in) :: na, n1
		integer                             :: i
		real                                :: a0, imp0

		call allocatehelix (imp, 1)	! implicit FD filter
		imp0       = b + eps		! zero lag coeff.
		imp%lag    = (/ 1 /)	! lag(1)= 1 # one side only
		imp%flt(1) = a1

		call allocatehelix (aa, na)	! implicit FD derivative
		aa%flt = 0.

		do i = 1, na  
			aa%lag (i     ) = i				! early lags (first row)
		end do 

		call wilson_init	(10 * n1)
		call wilson_factor	(20, imp0, imp, a0, aa)
		call wilson_close	()
		call deallocatehelix(imp)
	end function 

! Factor 1-D implicit FD scheme with kolmogoroff factorization.
	subroutine impfac1d_kol (b, a1, n1, na, f1, f2)
		real,				intent (in)		:: b, a1
		integer,			intent (in)		:: n1, na
		type (cfilter),		intent (out)	:: f1, f2
		integer								:: i
		type (cfilter)						:: spec

		call allocatechelix (spec, 3)
		call allocatechelix (f1, na)
		call allocatechelix (f2, na)

		spec%flt = (/a1, b, a1/) ! spectrum, zero lag coefficient in the center
		spec%lag = (/-1, 0, 1/)  ! lags

		f1%flt = (0. ,0.)
		f2%flt = (0. ,0.)

		do i = 1, na  
			f1%lag (i) = i-1	! create lags from 0 to na-1
			f2%lag (i) = i-1
		end do 

		call kolfactorchelix_alloc (16384) ! 16384 is the maximum padded length for which function xkolmog works

		call kolfactorchelix (spec, f1, f2)

		call kolfactorchelix_dealloc()
		call deallocatechelix (spec)
	end subroutine 

! Factor 2-D implicit FD 4th spatial order scheme, using kolmogoroff factorization.
	subroutine impfac4_kol (c, b1, b2, a1, a2, n1, na, f1, f2)
		real,			intent (in)		:: c, b1, b2, a1, a2
		integer,		intent (in)		:: n1, na
		type (cfilter),	intent (out)	:: f1, f2
		integer							:: i
		type (cfilter)					:: spec

		call allocatechelix (spec, 9)
		call allocatechelix (f1, 4*na + 1)
		call allocatechelix (f2, 4*na + 1)

		spec%flt = (/a2, b2, a1, b1, c, b1, a1, b2, a2/) ! spectrum, zero lag coefficient in the center
		spec%lag = (/-2*n1, -n1, -2, -1, 0, 1, 2, n1, 2*n1/)  ! lags

		f1%flt = (0. ,0.)
		f2%flt = (0. ,0.)

		do i = 0, na-1
			f1%lag (i + 1)			= i				! early lags (first row)
			f1%lag (na + i + 1)		= n1 + i - na	! middle lags (second row)
			f1%lag (2*na + i + 1)	= n1 + i		! middle lags (second row)
			f1%lag (3*na + i + 1)	= 2*n1 + i - na	! late lags (third row)

			f2%lag (i + 1)			= i 			! early lags (first row)
			f2%lag (na + i + 1)		= n1 + i - na	! middle lags (second row)
			f2%lag (2*na + i + 1)	= n1 + i 		! middle lags (second row)
			f2%lag (3*na + i + 1)	= 2*n1 + i - na	! late lags (third row)
		end do 

		f1%lag (4*na + 1) = 2*n1
		f2%lag (4*na + 1) = 2*n1


		call kolfactorchelix_alloc (16384) ! 16384 is the maximum padded length for which function xkolmog works

		call kolfactorchelix (spec, f1, f2)

		call kolfactorchelix_dealloc()
		call deallocatechelix (spec)
	end subroutine 

! convolution and correlation function for floats
	subroutine convolve (f, g, h, n, m)
		integer, intent(in)					:: n, m
		real, dimension(:), intent(in)	:: f, g
		real, dimension(:), intent(out)	:: h
		integer 							:: i, j

		h = 0.
	
		do i = 0, n-1
			do j = 0, m-1
				h (i+j+1) = h (i+j+1) + f(i+1) * g(j+1)
			end do
		end do
	end subroutine

	subroutine correlate (f, g, h, n)
		integer, intent(in)					:: n
		real, dimension(:), intent(in)	:: f, g
		real, dimension(:), intent(out)	:: h
		integer								:: i, j
	
		h = 0.
	
		do i = 0, n-1
			do j = i, n-1
				h (j-i+1) = h (j-i+1) + f(i+1) * g(j+1)
			end do
		end do
	end subroutine

! convolution and correlation function for doubles
	subroutine dconvolve (f, g, h, n, m)
		integer, intent(in)					:: n, m
		real(8), dimension(:), intent(in)	:: f, g
		real(8), dimension(:), intent(out)	:: h
		integer 							:: i, j

		h = 0.
	
		do i = 0, n-1
			do j = 0, m-1
				h (i+j+1) = h (i+j+1) + f(i+1) * g(j+1)
			end do
		end do
	end subroutine

	subroutine dcorrelate (f, g, h, n)
		integer, intent(in)					:: n
		real(8), dimension(:), intent(in)	:: f, g
		real(8), dimension(:), intent(out)	:: h
		integer								:: i, j
	
		h = 0.
	
		do i = 0, n-1
			do j = i, n-1
				h (j-i+1) = h (j-i+1) + f(i+1) * g(j+1)
			end do
		end do
	end subroutine

! convolution and correlation function for complex
	subroutine cconvolve (f, g, h, n, m)
		integer, intent(in)					:: n, m
		complex, dimension(:), intent(in)	:: f, g
		complex, dimension(:), intent(out)	:: h
		integer 							:: i, j

		h = (0., 0.)
	
		do i = 0, n-1
			do j = 0, m-1
				h (i+j+1) = h (i+j+1) + f(i+1) * conjg (g(j+1))
			end do
		end do
	end subroutine

	subroutine ccorrelate (f, g, h, n)
		integer, intent(in)					:: n
		complex, dimension(:), intent(in)	:: f, g
		complex, dimension(:), intent(out)	:: h
		integer								:: i, j
	
		h = (0., 0.)
	
		do i = 0, n-1
			do j = i, n-1
				h (j-i+1) = h (j-i+1) + f(i+1) * conjg(g(j+1))
			end do
		end do
	end subroutine

  !!---------------------------------------------------------------------------
  !!.. complex tridiagonal solver
  !!---------------------------------------------------------------------------
  !!.. a(n), sub-diagonal elemnets, only elements from 2:n are used
  !!.. b(n), diagonal elemetns
  !!.. c(n), super-diagonal elements, only elements from 1:n-1 are used
  !!
  !!   /  b(1), c(1)                                     \ / u(  1) \     / r(  1) \
  !!   |  a(2), b(2), c(2)                               | | u(  2) |     | r(  2) |
  !!   |        a(3), b(3), c(3)                         | | u(  3) |  =  | r(  3) |
  !!   |              ... ...                            | |  ...   |     |  ...   |
  !!   |                          a(n-1), b(n-1), c(n-1) | | u(n-1) |     | r(n-1) |
  !!   \                                  a(  n), b(  n) / \ u(  n) /     \ r(  n) /

	subroutine ctridiag_lab4 (a, b, c, r, u, n)
		integer            :: n
		real               :: a(n), b(n), c(n), r(n), u(n)
		integer            :: j
		integer, parameter :: NMAX = 5000
		real               :: bet, gam(NMAX)

		bet  = b(1)
		u(1) = r(1) / bet

		do j = 2, n-1
			gam(j) =  c(j-1) / bet
			bet    =  b(j) - a(j) * gam(j)
			u(j)   = (r(j) - a(j) * u(j-1)) / bet
		end do
		do j = n-1, 1, -1
			u(j)   = u(j) - gam(j+1) * u(j+1)
		end do
		return
	end subroutine

	subroutine rtris_lab7 (n, endl, a, b, c, endr, d, q)
		integer			:: i, n
		real			:: q(n), d(n), a, b, c, den, endl, endr
		real		 	:: f(n), e(n)
		
		e(1) = -a/endl
		f(1) = d(1)/endl
		
		do i= 2, n-1
			den = b+c*e(i-1)
			e(i) = -a/den
			f(i) = (d(i)-c*f(i-1))/den
		end do
		
		q(n) = (d(n)-c*f(n-1)) / (endr+c*e(n-1))
		
		do i= n-1, 1, -1
			q(i) = e(i) * q(i+1) + f(i)
		end do
		
		return
	end subroutine

	subroutine gauss_elim_square_2d (n, A, b)
		integer, intent(in)		:: n
		real, dimension(:,:)	:: A
		real, dimension(:)		:: b

		integer					:: i, j, k
		real					:: Upper, Lower, iUpper, iLower

		! zero lower half of matrix
		do k = 1, n-1	! for all columns / rows until row before last
			Upper  = A (k, k)
			iUpper = 1. / Upper
!write (0,10),  k, k
			do i = k+1, n	! for all rows below diagonal
				Lower = A (i, k)
!write (0,20), i, k
				b (i)    = (b (i) * Upper - b (k) * Lower) * iUpper
!write (0,30), i, k
				do j = k, n ! for columns in these rows
					A (i, j) = (A (i, j) * Upper - A (k, j) * Lower) * iUpper
!write (0,40), i, j, k, j
				end do

			end do
		end do

		! zero upper half of matrix
		do k = n, 2, -1		! for all columns / rows until second row
			Lower  = A (k, k)
			iLower = 1. / Lower
!write (0,20),  k, k
			do i = k-1, 1, -1	! for all rows above diagonal
				Upper = A(i ,k)
!write (0,10), i, k

				b (i) = (b (i) * Lower - b(k) * Upper) * iLower
!write (0,30), i, k

				do j = k, i, -1 ! for columns in these rows
					A (i, j) = (A(i, j) * Lower - A(k, j) * Upper) * iLower
!write (0,40), i, j, k, j
				end do
			end do
		end do

		! normalize by diagonal elements
		do i = 1, n
			if (abs(A(i,i)) .gt. 1e-10) then
				b (i)    = b (i) / A(i,i)
				A (i, i) = 1.
			end if
		end do

10	format ("Upper = A (", i2, ") (", i2, ")")
20	format ("Lower = A (", i2, ") (", i2, ")")
30	format ("b (", i2, ") * Upper - b (", i2, ") * Lower")
40	format ("A (", i2, ",", i2, ") * Upper - A (", i2, ",", i2, ") * Lower")

	end subroutine

	subroutine gauss_elim_square_1d (n, A, b)
		integer, intent(in)		:: n
		real, dimension(:)		:: A
		real, dimension(:)		:: b

		integer					:: i, j, k
		real					:: Upper, Lower, iLower, iUpper

		! zero lower half of matrix
		do k = 1, n-1	! for all columns / rows until row before last
			Upper  = A ((k-1) * n + k)
			iUpper = 1. / Upper
			do i = k+1, n	! for all rows below diagonal

				Lower = A ((i-1) * n + k)
				b (i)    = (b (i)    * Upper - b (k)    * Lower) * iUpper
				do j = k, n ! for columns in these rows

					A ((i-1) * n + j) = (A ((i-1) * n + j) * Upper - A ((k-1) * n + j) * Lower) * iUpper

				end do

			end do
		end do


		! zero upper half of matrix
		do k = n, 2, -1		! for all columns / rows until second row
			Lower  = A ((k-1) * n + k)
			iLower = 1. / Lower
			do i = k-1, 1, -1	! for all rows above diagonal
				Upper = A((i-1) * n + k)

				b (i) = (b (i) * Lower - b(k) * Upper) * iLower

				do j = k, i, -1 ! for columns in these rows
					A ((i-1) * n + j) = (A((i-1) * n + j) * Lower - A((k-1) * n + j) * Upper) * iLower
				end do

			end do
		end do

		! normalize by diagonal elements
		do i = 1, n
			if (abs (A((i-1) * n + i)) .gt. 1e-10) then
				b (i)			  = b (i) / A ((i-1) * n + i)
				A ((i-1) * n + i) = 1.
			end if
		end do
	end subroutine

	subroutine gauss_elim_offbands (n, n1, A, b)
		integer, intent(in)		:: n
		integer, intent(in)		:: n1
		real, dimension(:)		:: A
		real, dimension(:)		:: b

		integer					:: i, j, k
		real					:: Upper, Lower


		! zero band directly below diagonal
		do j = 2, n
			i = j-1
			k = i + n1

			Upper = A ((i-1) * n + i)
			Lower = A ((i-1) * n + j)

			A ((i-1) * n + j) = 0.
			A ((j-1) * n + j) = (A((j-1) * n + j) * Upper - A((j-1) * n + i) * Lower) / Upper
			A ((k-1) * n + j) = - A((k-1) * n + i) * Lower
			
			b (j) = (b (j) * Upper - b (i) * Lower) / Upper
		end do

		! zero far band below diagonal
		do j = 1 + n1, n
			i = j - n1

			Upper = A((i-1) * n + i)
			Lower = A((i-1) * n + j)

			A ((i-1) * n + j) = 0.
			A ((j-1) * n + j) = (A ((j-1) * n + j) * Upper - A((j-1) * n + i) * Lower) / Upper
			b (j) = (b (j) * Upper - b (i) * Lower) / Upper
		end do

		! zero band directly above diagonal
		do j = 1, n-1
			i = j+1

			Upper = A((i-1) * n + j)
			Lower = A((i-1) * n + i)

			A ((i-1) * n + j) = 0.
			b (j) = (b (j) * Lower - b (i) * Upper) / Lower
		end do

		! zero far band above diagonal
		do j = 1, n - n1
			i = j + n1

			Upper = A((i-1) * n + j)
			Lower = A((i-1) * n + i)

			A ((i-1) * n + j) = 0.
			b (j) = (b (j) * Lower - b (i) * Upper) / Lower
		end do

		! normalize by diagonal elements
		do i = 1, n
			if (abs (A((i-1) * n + i)) .gt. 1e-10) then
				b (i)			  = b (i) / A ((i-1) * n + i)
				A ((i-1) * n + i) = 1.
			end if
		end do

	end subroutine

end module prop2dutils


!-----------------
!COMMON Example

!INTEGER :: count, array(1000)
!COMMON /awful / count, array

!Make a file (say awful.f90) containing:

!MODULE awful
!INTEGER :: count, array(1000)
!END MODULE awful
!USE a
